/*
        alf: Piper program implementing and analysing traditional and historic c(ai)phers 
        designed to be used in conjuction with other programs.
        
        Copyright (C) 2004  Corcalciuc V. Horia

        This library is free software; you can redistribute it and/or
        modify it under the terms of the GNU Lesser General Public
        License as published by the Free Software Foundation; either
        version 2.1 of the License, or (at your option) any later version.
                
        This library is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
        Lesser General Public License for more details.
                                
        You should have received a copy of the GNU Lesser General Public
        License along with this library; if not, write to the Free Software
        Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

*/

/* Dedicated to Rudolf */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <locale.h>
#include <sys/time.h>
#include <math.h>
#include <popt.h>

enum Bool {
	false=0,
	true
} verboose=false, decrypt, check;

int _caesar(char *org);
int _atbash(char *org);
int _vigenere(char *org, char *key);
int _vernam(char *org, char *key);
int _playfair(char *org, char *key);

char *indata;

int main(const int argc, const char **argv) {

	poptContext optCon;
	char *optBuf;
	int result;
	int itra=0;
	char *key;
	char *store_checker;
	int sem=1;
	struct timeval *sop_te, *fop_te;
	struct timeval *sop_td, *fop_td;
	float fold, unfold;

	setvbuf(stdout, (char *) NULL, _IONBF, 0);
	setvbuf(stdin, (char *) NULL, _IONBF, 0);
	setlocale(LC_ALL, "");
	
	optBuf = calloc(argc, sizeof(char));
	indata = calloc(BUFSIZ, sizeof(char));
	key = calloc(BUFSIZ, sizeof(char));	

	const struct poptOption optionT[] = {
		{ "\b\b  \nCyphers:", 0, 0, NULL, 0, NULL},
		{ "caesar",	0, 0, NULL, 'c', "CAESAR", NULL},
		{ "atbash", 0, 0, NULL, 'b', "ATBASH", NULL},
		{ "vigenere", 0, POPT_ARG_STRING, &key, 'v', "VIGENERE", "KEY"},
		{ "vernam", 0, POPT_ARG_STRING, &key, 'k', "VERNAM", "FILE"},
		{ "playfair", 0, POPT_ARG_STRING, &key, 'f', "PLAYFAIR", "KEY"},
		{ "\b\b  \nOptional:", 0, 0, NULL, 0, NULL},
		{ NULL, 'c', 0, NULL, 'a', "Efficiency checks", NULL},
		{ "\b\b  \nMandatory:", 0, 0, NULL, 0, NULL},
		{ NULL, 'd', 0, NULL, 'd', "Decrypt text", NULL},
		{ NULL, 'x', 0, NULL, 'x', "Encrypt text", NULL},
		{ "\b\b  \nHelp:", 0, 0, NULL, 0, NULL },
		{ "help", 'h', 0, NULL, 'h', "Show help message", NULL},
		{NULL, 0, 0, NULL, 0}
	};
	
	optCon = poptGetContext(NULL, argc, argv, optionT, 0);

		while((result = poptGetNextOpt(optCon)) >=0)
		{
			switch(result)
			{
				case 'h':
				optBuf[itra++]='h';
				break;
				case 'c':
				optBuf[itra++]='c';
				break;
				case 'b':
				optBuf[itra++]='b';
				break;
				case 'v':
				optBuf[itra++]='v';		
				break;
				case 'k':
				optBuf[itra++]='k';
				break;
				case 'f':
				optBuf[itra++]='f';
				break;
				case 'a':
				optBuf[itra++]='a';
				break;
				case 'd':
				optBuf[itra++]='d';
				break;
				case 'x':
				optBuf[itra++]='x';
				break;
			}
		}
		
		if (result < -1)
			fprintf (stderr, "%s: %s\n", poptBadOption (optCon, POPT_BADOPTION_NOALIAS), poptStrerror (result)), exit(1);
		else if (argc <2)
			fprintf (stderr, "%s: please select a cypher\n", argv[0]), exit(1);

		if ( strchr(optBuf, 'h') != NULL ) {
			poptPrintHelp(optCon, stderr, 0);
			exit(0);
		}	
	
		for(itra=0; itra<=BUFSIZ; itra++)
			if(!feof(stdin))
				indata[itra]=fgetc(stdin);
			else {
				indata[itra-1]='\0';
				break;
			}
	
	if ( strchr(optBuf, 'd') != NULL )
		{
			decrypt=true;
		}
	else if (strchr(optBuf, 'x') != NULL )
		{
			decrypt=false;
		}
	else if (strchr(optBuf, 'a') != NULL )
		{
			check=true;
			decrypt=false;
			store_checker=calloc(strlen(indata), sizeof(char));
			strcpy(store_checker, indata);
			sop_te=calloc(1, sizeof(struct timeval));
			fop_te=calloc(1, sizeof(struct timeval));
			sop_td=calloc(1, sizeof(struct timeval));
			fop_td=calloc(1, sizeof(struct timeval));
		}
	else
		{
			fprintf(stderr, "%s: please select encryption, decryption or check.\n", argv[0]), exit(1);
		}
		
sequence:
	{
			if(sem && check) gettimeofday(sop_te, 0);
								
			if ( strchr (optBuf, 'c') != NULL )
				{
					_caesar ( indata );
				}
			if ( strchr(optBuf, 'b') != NULL )
				{
					_atbash ( indata );
				}
			if ( strchr(optBuf, 'v') != NULL )
				{
					_vigenere ( indata, key );
				}
			if ( strchr(optBuf, 'k') != NULL )
				{
					_vernam ( indata, key );
				}
			if ( strchr(optBuf, 'f') != NULL )
				{
					_playfair ( indata, key );
				}
			
			if(!sem && check) goto checks; 
				else gettimeofday(fop_te, 0);
	}
		
	if(!check)
		fprintf(stdout, "%s", indata);
	else if(check)
		{
			gettimeofday(sop_td, 0);
			sem--;
			decrypt=true;
			goto sequence;

checks:
	{
	
			gettimeofday(fop_td, 0);
			fprintf(stdout, "Fold time: %0.3fms\n", fold=((fop_te->tv_sec<=sop_te->tv_usec) ? (float) (fop_te->tv_sec-sop_te->tv_sec) * 1000 : (fop_te->tv_usec-sop_te->tv_usec)*0.001));
			fprintf(stdout, "Unfold time: %0.3fms\n", unfold=((fop_td->tv_sec<=sop_td->tv_usec) ? (float) (fop_td->tv_sec-sop_td->tv_sec) * 1000 : (float) (fop_td->tv_usec-sop_td->tv_usec)*0.001));
	
			fprintf(stdout, "Consistency: ");
			if(!strncmp(store_checker, indata, strlen(indata)))
				fprintf(stdout, "100%%\n");
			else {
				for(result=itra, itra=0; itra<=strlen(indata); itra++, srand((unsigned int) time(NULL)))
					if(indata[itra]!=store_checker[itra])
						result--;
					else if(!isgraph(indata[itra]))
						result--;
					else if(isspace(indata[itra]) && isspace(indata[itra+rand()%2]))
						result--;
					else if(ispunct(indata[itra]) && (isalpha(indata[itra-rand()%2]) || isalpha(indata[itra+rand()%2])))
						result--;
				fprintf(stdout, "%.2f%%\n", ((float) result/ (float) strlen(indata)) *100);
			}
			
			fprintf(stdout, "Armor: ");
			for(itra=0, result=0; itra<strlen(store_checker); itra++)
				switch(store_checker[itra])
					{
						case '\n':
						case ':':
								result++;
								break;
						case ' ':
						case '\"':
						case '\'':
						case ',':
						case '.':
								result+=2;
								break;
						case '!':
						case '?':
								result+=3;
								break;
					};
			fprintf(stdout, "%.4f\n", (double) (unfold/fold)/(double) itra*result);
				
	}
	
		}
	
	poptFreeContext(optCon);
	exit(0);
}

int _caesar(char *org) {

	char *dst;
	int itra;
	
	dst = (char *) malloc(strlen(org));
	
	for ( itra=0; itra<=strlen(org); itra++ )
		{
			if (!isalpha(org[itra]))
				dst[itra] = org[itra];
			else {
				if (!decrypt)
					dst[itra] = org[itra]+3;
				else
					dst[itra] = org[itra]-3;
			}			
		}
	
	indata=calloc(strlen(dst),sizeof(char));
	strncpy(indata, dst, strlen(dst));
	free(dst);
	
	return 0;
}

int _atbash(char *org) {

	char *dst;
	int itra;
	int sem=1;
	
	dst = calloc (strlen(org), sizeof(char));
	
	for ( itra = 0; itra<=strlen(org); itra++ )
		{
			if (isupper(org[itra])) {
				org[itra]=tolower((char) org[itra]);
				sem--;
			}
			if (!isalpha(org[itra]))
				dst[itra]=org[itra];
			else {
					dst[itra] = 'z' - ( org[itra] - 'a' );
					if(!sem) {
						dst[itra]=toupper((char ) dst[itra]);
						sem++;
					}
			}
		}

	indata=calloc(strlen(dst),sizeof(char));
	strncpy(indata, dst, strlen(dst));
	free(dst);
		
	return 0;

}

int _vigenere(char *org, char *key) {

	int itra, itrb, itrc, pad, sem=1;
	char *hlpr[strlen(key)];
	char *dst;
	
	for(itra=0; itra<strlen(key); itra++)
		if ( isdigit(key[itra]) )
			return 1;

	dst = calloc (strlen(org), sizeof(char));

	for (itra=0; itra<strlen(key); itra++)
		{
			hlpr[itra] = (char *) malloc('z'-'a');
			for (itrb=0, pad=0; itrb<=25; itrb++)
				{
					if (key[itra]+pad>'z')
						pad=('a'-key[itra]);
					hlpr[itra][itrb]=key[itra]+(pad++);
				}
		}
	
	strcpy(dst, org);
	
	for (itra=0, itrc=0; itra<strlen(org)-1; itra++) {
		if(itrc>=strlen(key))
			itrc=0;
		if( isupper(org[itra]) ) {
			org[itra]=tolower(org[itra]);
			sem--;
		}
		if( !isalpha(org[itra]) ) {
				dst[itra]=org[itra];
				itrc=0;
			}
		else 
			{
				if(!decrypt)
					dst[itra]=hlpr[itrc++][org[itra]-'a'];
				else if(decrypt) {
					for(itrb=0; itrb<strlen(hlpr[itrc]); itrb++) {
						if(hlpr[itrc][itrb]==org[itra])
							dst[itra]=itrb+'a';
					}
					itrc++;
				}
				if(!sem){
					dst[itra]=toupper(dst[itra]);
					sem++;
				}
			}
	}
	
	indata=calloc(strlen(dst), sizeof(char));
	strncpy(indata, dst, strlen(dst));
	free(dst);
		
	return 0;	
}

int _vernam(char *org, char *key)
{

	int itra=0, sem=1, pad=0; 
	char c, *hlpr, *dst;
	FILE *fp;
	
	dst=calloc(strlen(org), sizeof(char));
	hlpr=calloc(strlen(org), sizeof(char));

	if((fp=fopen(key, "r"))==NULL)
		fprintf(stderr, "%s: %s", key, strerror(errno)),exit(0);

	while((c=fgetc(fp))!=EOF && itra<strlen(org)-1)
		if( isalpha(c) )
			hlpr[itra++]=c;
				
	if(strlen(org)-1>strlen(hlpr))
		fprintf(stderr, "%s: %s\n", "alf", "Entropy too short!"), exit(1);
	
	for(itra=0; itra<strlen(org)+strlen(key)+1; itra++) 
		{
		
			if( isupper(org[itra]) )
				{
					org[itra]=tolower(org[itra]);
					sem--;
				}
			if( !isalpha(org[itra]) )
					dst[itra]=org[itra];
			else if(!decrypt)
				{
					pad='a'+(org[itra]-'a')+(hlpr[itra]-'a');
					if(pad>'z')
						pad-=26;
					dst[itra]=pad;
				}
			else if(decrypt)
				{
					pad='a'+(org[itra]-'a')-(hlpr[itra]-'a');
						if(pad<'a')
							pad+=26;
						dst[itra]=pad;
				}
			if(!sem){
					dst[itra]=toupper(dst[itra]);
					sem++;
				}
		}
	
	if(strchr(org, '\n')!=NULL)
		dst[strlen(dst)]='\n';
			
	indata=calloc(strlen(dst), sizeof(char));
	strncpy(indata, dst, strlen(dst));
	free(dst);
		
	return 0;	

}

int _playfair(char *org, char *key)
{
	int lph='a', pad=0, sem=1, itra=0;
	char *dst, *didst, *hlpr[5], *fidst;
	char *n_key;
	
	struct coords {
		int x;
		int y;
		int xs;
		int ys;
		int xf;
		int yf;
		int xfs;
		int yfs;
	} *loc;

	dst=calloc(strlen(org), sizeof(char));
	n_key=calloc(strlen(key), sizeof(char));
	loc=calloc(1, sizeof(struct coords));

	for(loc->x=0; loc->x<5; loc->x++)
		hlpr[loc->x]=calloc(5, sizeof(char));		
	
	for(loc->x=0, pad=0; loc->x<=strlen(key); loc->x++) {
		if(strchr(n_key, key[loc->x])==NULL)
				n_key[pad++]=key[loc->x];
	}
	
	for(loc->y=0, pad=0; loc->y<5 && pad<strlen(n_key); loc->y++)
		for(loc->x=0; loc->x<5 && pad<strlen(n_key); loc->x++)
			hlpr[loc->y][loc->x]=n_key[pad++];

	for(sem=loc->y-1; lph<='z' && loc->y<6; lph++)	{
		if(lph=='q')
			continue;
		if(loc->x>4) {
			loc->x=0;
			loc->y++;
		}
		for(pad=0; pad<=sem; pad++)
			if(strchr(hlpr[pad], lph)!=NULL)
					break;
		if(pad>sem)
				hlpr[loc->y-1][loc->x++]=lph;
	}
	
	if(verboose==true)
		for(pad=0; pad<5; pad++)
			fprintf(stdout, " %c %c %c %c %c\n", hlpr[pad][0], hlpr[pad][1], hlpr[pad][2], hlpr[pad][3], hlpr[pad][4]);
	
	didst=calloc(2*strlen(org), sizeof(char));
	for(itra=0, pad=0, loc->x=0; itra<strlen(org); itra++, pad++) {
		if(isalpha(org[itra]))
			loc->x++;
		didst[pad]=org[itra];
		didst[++pad]=' ';
	}
	
	for(pad=0; pad<strlen(didst); pad+=2) {
		if(didst[pad+2]==didst[pad]) {
			didst[pad+1]='x';
			loc->x++;
		}
	}
	
	fidst=calloc(10, sizeof(char));
	for(pad=0, itra=0; pad<strlen(didst); pad++)
		if(isalpha(didst[pad]))
			fidst[itra++]=didst[pad];

	free(didst);
	if(verboose==true)
		fprintf(stdout, "didst: %s\n", fidst);

	for(pad=0, sem=0, itra=0; pad<=strlen(org); pad++, sem++)
		{
			for(loc->y=0; loc->y<5; loc->y++)
				for(loc->x=0; loc->x<5; loc->x++)
					if(hlpr[loc->y][loc->x]==fidst[sem]) {
						loc->xs=loc->x;
						loc->ys=loc->y;
					}
			for(loc->y=0; loc->y<5; loc->y++)
				for(loc->x=0; loc->x<5; loc->x++)
					if(hlpr[loc->y][loc->x]==fidst[sem+1]) {
					
					if(verboose==true)
						fprintf(stdout, "%c:x%d-y%d %c:x%ds-y%ds\n", fidst[sem], loc->xs, loc->ys, fidst[sem+1], loc->x, loc->y);
						
						if(loc->ys==loc->y) {
							if(loc->xs+1>4)
								loc->xfs=0;
							else
								loc->xfs=loc->xs+1;
							if(loc->x+1>4)
								loc->xf=0;
							else
								loc->xf=loc->x+1;
						dst[itra++]=hlpr[loc->y][loc->xfs];
						dst[itra++]=hlpr[loc->y][loc->xf];
						}
						if(loc->xs==loc->x) {
							if(loc->ys+1>4)
								loc->yfs=0;
							else
								loc->yfs=loc->ys+1;
							if(loc->y+1>4)
								loc->yf=0;
							else
								loc->yf=loc->y+1;
						dst[itra++]=hlpr[loc->yfs][loc->x];
						dst[itra++]=hlpr[loc->yf][loc->x];
						}
						if(loc->xs!=loc->x && loc->ys!=loc->y) {
							if(loc->xs>loc->x) {
								loc->xf=loc->x;
								loc->yf=loc->ys;
								loc->xfs=loc->xs;
								loc->yfs=loc->y;
							}
							else if(loc->xs<loc->x) {
								loc->xf=loc->x;
								loc->yf=loc->ys;
								loc->xfs=loc->xs;
								loc->yfs=loc->y;
							}
						else if(loc->xs==loc->y) {
								loc->xfs=loc->xs;
								loc->yfs=loc->y;
								loc->xf=loc->x;
								loc->yf=loc->ys;
						} else if(verboose==true)
								fprintf(stdout, "Black zone\n");
								dst[itra++]=hlpr[loc->yf][loc->xf];
								dst[itra++]=hlpr[loc->yfs][loc->xfs];
						}
					}
			sem++;
		}

	if(strchr(org, '\n')!=NULL)
		dst[strlen(dst)]='\n';

	indata=calloc(strlen(dst), sizeof(char));
	strncpy(indata, dst, strlen(dst));
	free(dst);
	return 0; 
	
}
